@@ -5,7 +5,7 @@
 ;; Author: Wojciech Siewierski
 ;;         Rudolf Schlatte <rudi@constantly.at>
 ;; URL: https://codeberg.org/emacs-jj-vc/vc-jj.el
-;; Version: 0.2
+;; Version: 0.3
 ;; Package-Requires: ((emacs "28.1") (compat "29.4"))
 ;; Keywords: vc tools
 
@@ -52,18 +52,34 @@
   :type 'string
   :risky t)
 
-(defcustom vc-jj-colorize-log t
-  "Control whether to have jj colorize the log."
-  :type 'boolean)
-
-(defcustom vc-jj-log-template "builtin_log_compact"
-  "The template to use for `vc-print-log'."
-  :type '(radio (const "builtin_log_oneline")
-                (const "builtin_log_compact")
-                (const "builtin_log_comfortable")
-                (const "builtin_log_compact_full_description")
-                (const "builtin_log_detailed")
-                (string :tag "Custom template")))
+(defvar vc-jj--log-default-template
+  "
+if(root,
+  format_root_commit(self),
+  label(if(current_working_copy, \"working_copy\"),
+    concat(
+      separate(\" \",
+        change_id.shortest(8).prefix() ++ \"​\" ++ change_id.shortest(8).rest(),
+        if(author.name(), author.name(), if(author.email(), author.email().local(), email_placeholder)),
+        commit_timestamp(self).format(\"%Y-%m-%d\"),
+        bookmarks,
+        tags,
+        working_copies,
+        if(git_head, label(\"git_head\", \"git_head()\")),
+        format_short_commit_id(commit_id),
+        if(conflict, label(\"conflict\", \"conflict\")),
+        if(config(\"ui.show-cryptographic-signatures\").as_boolean(),
+          format_short_cryptographic_signature(signature)),
+        if(empty, label(\"empty\", \"(empty)\")),
+        if(description,
+          description.first_line(),
+          label(if(empty, \"empty\"), description_placeholder),
+        ),
+      ) ++ \"\n\",
+    ),
+  )
+)
+")
 
 (defcustom vc-jj-global-switches '("--no-pager" "--color" "never")
   "Global switches to pass to any jj command."
@@ -97,40 +113,64 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 (defun vc-jj--filename-to-fileset (filename &optional root)
   "Convert FILENAME to a JJ fileset expression.
-If ROOT is given, the result is relative to the repository root,
-otherwise it is relative to the current directory."
+If ROOT is non-nil, the resulting fileset is relative to the repository
+root, otherwise it is relative to the current directory."
   (if root
       (concat "root-file:\""
-          (string-replace "\"" "\\\"" (file-relative-name filename root))
-          "\"")
+              (string-replace "\"" "\\\"" (file-relative-name filename root))
+              "\"")
     (concat "cwd-file:\""
             (string-replace "\"" "\\\"" (file-relative-name filename)) "\"")))
 
-(defun vc-jj-command (buffer okstatus file-or-list &rest flags)
+(defun vc-jj--process-lines (&rest args)
+  "Run jj with ARGS, returning its output to stdout as a list of strings.
+In contrast to `process-lines', discard output to stderr since jj prints
+warnings to stderr even when run with '--quiet'."
+  (with-temp-buffer
+    (let ((status (apply #'call-process vc-jj-program nil
+                         ;; (current-buffer)
+                         (list (current-buffer) nil)
+                         nil args)))
+      (unless (eq status 0)
+	(error "'jj' exited with status %s" status))
+      (goto-char (point-min))
+      (let (lines)
+	(while (not (eobp))
+	  (setq lines (cons (buffer-substring-no-properties
+			     (line-beginning-position)
+			     (line-end-position))
+			    lines))
+	  (forward-line 1))
+	(nreverse lines)))))
+
+(defun vc-jj--command-parseable (&rest args)
+  "Run jj with ARGS, returning its output as string.
+
+Note: any filenames in ARGS should be converted via
+`vc-jj--filename-to-fileset'.
+
+In contrast to `vc-jj--command-dispatched', discard output to stderr so
+the output can be safely parsed.  Does not support many of the features
+of `vc-jj--command-dispatched', such as async execution and checking of
+process status."
+  (with-temp-buffer
+    (let ((status (apply #'call-process vc-jj-program nil
+                         (list (current-buffer) nil)
+                         nil args)))
+      (unless (eq status 0)
+	(error "'jj' exited with status %s" status))
+      (buffer-substring-no-properties (point-min) (point-max)))))
+
+(defun vc-jj--command-dispatched (buffer okstatus file-or-list &rest flags)
   "Execute `vc-jj-program', notifying the user and checking for errors.
+See `vc-do-command' for documentation of the BUFFER, OKSTATUS,
+FILE-OR-LIST and FLAGS arguments.
 
-The output goes to BUFFER, the current buffer if BUFFER is t, or a
-buffer named \"*vc*\" if BUFFER is nil.  If the destination buffer is
-not already current, set it up properly and erase it.
-
-The command is considered successful if its exit status does not exceed
-OKSTATUS (if OKSTATUS is nil, that means to ignore error status, if it
-is 'async', that means not to wait for termination of the subprocess; if
-it is t it means to ignore all execution errors).  On unsuccessful
-execution, raise an error.
-
-FILE-OR-LIST is the name of a working file; it may be a list of files or
-be nil (to execute commands that don't expect a file name or set of
-files).  If an optional list of FLAGS is present, that is inserted into
-the command line before the filename(s).
-
-Return the return value of the command in the synchronous case, and the
-process object in the asynchronous case."
-  (let* ((filesets (mapcar #'vc-jj--filename-to-fileset
-                           (if (listp file-or-list) file-or-list (list file-or-list))))
-         (global-switches (if (stringp vc-jj-global-switches)
-                              (list vc-jj-global-switches)
-                            vc-jj-global-switches)))
+Note: if we need to parse jj's output `vc-jj--command-parseable' should
+be used instead of this function, since jj might print warnings to
+stderr and1 `vc-do-command' cannot separate output to stdout and stderr."
+  (let* ((filesets (mapcar #'vc-jj--filename-to-fileset (ensure-list file-or-list)))
+         (global-switches (ensure-list vc-jj-global-switches)))
     (apply #'vc-do-command (or buffer "*vc*") okstatus vc-jj-program
            ;; Note that we pass NIL for FILE-OR-LIST to avoid
            ;; vc-do-command mangling of filenames; we pass the fileset
@@ -166,12 +206,11 @@ process object in the asynchronous case."
   ;;   "M ") or added (output starts with "A "), but no output could
   ;;   be conflicted, ignored or unchanged
   (let* ((default-directory (vc-jj-root file))
+         (file (vc-jj--filename-to-fileset file))
          (conflicted-ignored
-          (with-output-to-string
-            (vc-jj-command standard-output 0 file "file" "list" "-T" "conflict" "--")))
+          (vc-jj--command-parseable "file" "list" "-T" "conflict" "--" file))
          (modified-added
-          (with-output-to-string (vc-jj-command standard-output 0 file
-                                                "diff" "--summary" "--"))))
+          (vc-jj--command-parseable "diff" "--summary" "--" file)))
     (cond
      ((string-empty-p conflicted-ignored) 'ignored)
      ((string= conflicted-ignored "true") 'conflict)
@@ -203,11 +242,11 @@ Return the result result of applying UPDATE-FUNCTION to that list."
   (let* ((dir (expand-file-name dir))
          (default-directory dir)
          (project-root (vc-jj-root dir))
-         (registered-files (process-lines vc-jj-program "file" "list" "--" dir))
+         (registered-files (vc-jj--process-lines "file" "list" "--" dir))
          (ignored-files (seq-difference (cl-delete-if #'file-directory-p
                                                       (directory-files dir nil nil t))
                                         registered-files))
-         (changed (process-lines vc-jj-program "diff" "--summary" "--" dir))
+         (changed (vc-jj--process-lines "diff" "--summary" "--" dir))
          (added-files (mapcan (lambda (entry)
                                 (and (string-prefix-p "A " entry)
                                      (list (substring entry 2))))
@@ -221,9 +260,8 @@ Return the result result of applying UPDATE-FUNCTION to that list."
          ;; expand-file-name / file-relative-name
          (conflicted-files (mapcar (lambda (entry)
                                      (file-relative-name (expand-file-name entry project-root) dir))
-                                   (process-lines vc-jj-program
-                                                  "file" "list"
-                                                  "-T" "if(conflict, path ++ \"\\n\")" "--" dir)))
+                                   (vc-jj--process-lines "file" "list"
+                                                         "-T" "if(conflict, path ++ \"\\n\")" "--" dir)))
          (unchanged-files (cl-remove-if (lambda (entry) (or (member entry conflicted-files)
                                                             (member entry modified-files)
                                                             (member entry added-files)
@@ -247,64 +285,96 @@ current change is conflicted, divergent, hidden or immutable, also add a
 Status header.  (We do not check for emptiness of the current change
 since the user can see that via the list of files below the headers
 anyway.)"
-  (pcase-let* ((default-directory (file-name-as-directory dir))
-               (`( ,change-id ,change-id-short ,commit-id ,commit-id-short
-                   ,description ,bookmarks ,conflict ,divergent ,hidden ,immutable)
-                (process-lines vc-jj-program "log" "--no-graph" "-r" "@" "-T"
-                               "concat(
-self.change_id().short(), \"\\n\",
+  (cl-destructuring-bind
+      ;; We have some fixed lines for the current changeset, then at
+      ;; the end 5 lines for each parent
+      ( change-id change-id-short commit-id commit-id-short
+        description bookmarks conflict divergent hidden immutable
+        &rest parent-info)
+      (let ((default-directory (file-name-as-directory dir)))
+        (vc-jj--process-lines "log" "--no-graph" "-r" "@" "-T"
+                              "concat(
+self.change_id().short(8), \"\\n\",
 self.change_id().shortest(), \"\\n\",
-self.commit_id().short(), \"\\n\",
+self.commit_id().short(8), \"\\n\",
 self.commit_id().shortest(), \"\\n\",
 description.first_line(), \"\\n\",
 bookmarks.join(\",\"), \"\\n\",
 self.conflict(), \"\\n\",
 self.divergent(), \"\\n\",
 self.hidden(), \"\\n\",
-self.immutable(), \"\\n\"
-)"))
-               (status (concat
-                        (and (string= conflict "true") "(conflict)")
-                        (and (string= divergent "true") "(divergent)")
-                        (and (string= hidden "true") "(hidden)")
-                        (and (string= immutable "true") "(immutable)")))
-               (change-id-suffix (substring change-id (length change-id-short)))
-               (commit-id-suffix (substring commit-id (length commit-id-short))))
-    (cl-flet ((fmt (key value &optional prefix)
-                (concat
-                 (propertize (format "% -11s: " key) 'face 'vc-dir-header)
-                 ;; there is no header value emphasis face, so we use
-                 ;; vc-dir-status-up-to-date for the prefix.
-                 (and prefix (propertize prefix 'face 'vc-dir-status-up-to-date))
-                 (propertize value 'face 'vc-dir-header-value))))
-      (string-join (seq-remove
-                    #'null
-                    (list
-                     (fmt "Description" (if (string= description "") "(no description set)" description))
-                     (fmt "Change ID" change-id-suffix change-id-short)
-                     (fmt "Commit" commit-id-suffix commit-id-short)
-                     (unless (string= bookmarks "") (fmt "Bookmarks" bookmarks))
-                     (unless (string= status "")
-                       ;; open-code this line instead of adding a
-                       ;; `face' parameter to `fmt'
-                       (concat
-                        (propertize (format "% -11s: " "Status") 'face 'vc-dir-header)
-                        (propertize status 'face 'vc-dir-status-warning)))))
-                   "\n"))))
+self.immutable(), \"\\n\",
+parents.map(|c| concat(
+  c.change_id().short(8), \"\\n\",
+  c.change_id().shortest(), \"\\n\",
+  c.commit_id().short(8), \"\\n\",
+  c.commit_id().shortest(), \"\\n\",
+  c.description().first_line(), \"\\n\"
+)))"))
+    (cl-labels
+        ((str (string &optional face prefix)
+           ;; format a string
+           (cond ((not face) (propertize string 'face 'vc-dir-header-value))
+                 ((not prefix) (propertize string 'face face))
+                 (t (concat (propertize prefix 'face 'vc-dir-header-value)
+                            (propertize string 'face face)))))
+         (info (key description change-id change-id-prefix commit-id commit-id-prefix)
+           ;; format a changeset info line
+           (let ((change-id-suffix (substring change-id (length change-id-prefix)))
+                 (commit-id-suffix (substring commit-id (length commit-id-prefix))))
+             (concat
+              (str (format "% -11s: " key) 'vc-dir-header)
+              ;; There's no vc-dir-header-value-emphasis or similar
+              ;; face, so we re-use vc-dir-status-up-to-date to render
+              ;; the unique prefix
+              " "
+              (str change-id-suffix 'vc-dir-status-ignored change-id-prefix)
+              " "
+              (str commit-id-suffix 'vc-dir-status-ignored commit-id-prefix)
+              " "
+              (if (string-empty-p description)
+                  (str "(no description set)")
+                (str description))))))
+      (let ((status (concat
+                     (and (string= conflict "true") "(conflict)")
+                     (and (string= divergent "true") "(divergent)")
+                     (and (string= hidden "true") "(hidden)")
+                     (and (string= immutable "true") "(immutable)")))
+            (parent-keys (cl-loop
+                          for (change-id change-id-short commit-id commit-id-short description)
+                          in (seq-partition parent-info 5)
+                          collect (info "Parent" description
+                                        change-id change-id-short
+                                        commit-id commit-id-short))))
+        (string-join
+         (seq-remove
+          ;; Remove NIL entries because we get empty lines otherwise
+          #'null
+          (cl-list*
+           (info "Changeset" description change-id change-id-short commit-id commit-id-short)
+           (unless (string= bookmarks "")
+             (concat (str "Bookmarks  : " 'vc-dir-header) (str bookmarks)))
+           (unless (string= status "")
+             (concat (str "Status     : " 'vc-dir-header) (str status 'vc-dir-status-warning)))
+           parent-keys))
+         "\n")))))
 
 (defun vc-jj-working-revision (file)
   "Return the current change id of the repository containing FILE."
   (when-let* ((default-directory (vc-jj-root file)))
-    (car (process-lines vc-jj-program "log" "--no-graph"
-                        "-r" "@"
-                        "-T" "change_id ++ \"\\n\""))))
+    ;; 'jj log' might print a warning at the start of its output,
+    ;; e.g., "Warning: Refused to snapshot some files".  The output we
+    ;; want will be printed afterwards.
+    (car (last (vc-jj--process-lines "log" "--no-graph"
+                                     "-r" "@"
+                                     "-T" "change_id ++ \"\\n\"")))))
 
 (defun vc-jj-mode-line-string (file)
   "Return a mode line string and tooltip for FILE."
   (pcase-let* ((long-rev (vc-jj-working-revision file))
                (`(,short-rev ,description)
-                (process-lines vc-jj-program "log" "--no-graph" "-r" long-rev
-                               "-T" "self.change_id().shortest() ++ \"\\n\" ++ description.first_line() ++ \"\\n\""))
+                (vc-jj--process-lines "log" "--no-graph" "-r" long-rev
+                                      "-T" "self.change_id().shortest() ++ \"\\n\" ++ description.first_line() ++ \"\\n\""))
                (def-ml (vc-default-mode-line-string 'JJ file))
                (help-echo (get-text-property 0 'help-echo def-ml))
                (face   (get-text-property 0 'face def-ml)))
@@ -329,28 +399,31 @@ self.immutable(), \"\\n\"
   ;; run "jj file track" for the case where some of FILES are excluded
   ;; via the "snapshot.auto-track" setting or via git's mechanisms
   ;; such as the .gitignore file.
-  (vc-jj-command nil 0 files "file" "track" "--"))
+  (vc-jj--command-dispatched nil 0 files "file" "track" "--"))
 
 (defun vc-jj-delete-file (file)
   "Delete FILE and make sure jj registers the change."
   (when (file-exists-p file)
     (delete-file file)
-    (vc-jj-command nil 0 nil "status")))
+    (vc-jj--command-dispatched nil 0 nil "status")))
 
 (defun vc-jj-rename-file (old new)
   "Rename file OLD to NEW and make sure jj registers the change."
   (rename-file old new)
-  (vc-jj-command nil 0 nil "status"))
+  (vc-jj--command-dispatched nil 0 nil "status"))
 
 (defun vc-jj-checkin (files comment &optional _rev)
   "Run \"jj commit\" with supplied FILES and COMMENT."
   (setq comment (replace-regexp-in-string "\\`Summary: " "" comment))
   (let ((args (append (vc-switches 'jj 'checkin) (list "commit" "-m" comment))))
-    (apply #'vc-jj-command nil 0 files args)))
+    (apply #'vc-jj--command-dispatched nil 0 files args)))
 
 (defun vc-jj-find-revision (file rev buffer)
   "Read revision REV of FILE into BUFFER and return the buffer."
-  (vc-jj-command buffer 0 file "file" "show" "-r" rev "--")
+  (let ((revision (vc-jj--command-parseable "file" "show" "-r" rev "--" (vc-jj--filename-to-fileset file))))
+    (with-current-buffer buffer
+      (erase-buffer buffer)
+      (insert revision)))
   buffer)
 
 (defun vc-jj-checkout (file &optional rev)
@@ -359,31 +432,31 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
   ;; TODO: check that this does the right thing: if REV is not
   ;; specified, should we revert or leave FILE unchanged?
   (let ((args (append (and rev (list "--from" rev)))))
-    (apply #'vc-jj-command nil 0 file "restore" args)))
+    (apply #'vc-jj--command-dispatched nil 0 file "restore" args)))
 
 (defun vc-jj-revert (file &optional _contents-done)
   "Restore FILE to the state from its parent(s), via \"jj restore\"."
-  (vc-jj-command nil 0 file "restore"))
+  (vc-jj--command-dispatched nil 0 file "restore"))
 
 (defun vc-jj-print-log (files buffer &optional _shortlog start-revision limit)
   "Print commit log associated with FILES into specified BUFFER."
   ;; FIXME: limit can be a revision string, in which case we should
   ;; print revisions between start-revision and limit
+  (vc-setup-buffer buffer)
   (let ((inhibit-read-only t)
         (args (append
                (and limit
                     (list "-n" (number-to-string limit)))
-               (and start-revision
-                    (list "-r" (concat ".." start-revision)))
-               (and vc-jj-colorize-log (list "--color" "always"))
-               (list "-T" vc-jj-log-template "--")
-               files)))
-    (with-current-buffer buffer (erase-buffer))
-    (apply #'call-process vc-jj-program nil buffer nil "log" args)
-    (when vc-jj-colorize-log
-      (with-current-buffer buffer
-        (ansi-color-apply-on-region (point-min) (point-max)))))
-  (goto-char (point-min)))
+               (if start-revision
+                 (list "-r" (concat "::" start-revision))
+                 (list "-r" "::"))
+               (list "-T" vc-jj--log-default-template "--")
+               (unless (string-equal (vc-jj-root (car files)) (car files))
+                 files))))
+    (with-current-buffer buffer
+      (apply #'vc-jj--command-dispatched buffer
+        'async nil
+        args))))
 
 (defun vc-jj-show-log-entry (revision)
   "Move to the log entry for REVISION."
@@ -403,6 +476,153 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
 ;;   ;; TODO
 ;;   )
 
+(defvar log-view-message-re)
+(defvar log-view-file-re)
+(defvar log-view-font-lock-keywords)
+(defvar log-view-per-file-logs)
+(defvar log-view-expanded-log-entry-function)
+
+(defvar vc-jj--logline-re
+  (rx
+    line-start
+    ;; graph
+    (+? nonl)
+    " "
+    ;; change-id
+    (group (+ (any "K-Zk-z")))
+    space
+    (group (+ (any "K-Zk-z")))
+    " "
+    ;; author
+    (group (* nonl))
+    " "
+    ;; time
+    (group
+      (= 4 (any num)) "-" (= 2 (any num)) "-" (= 2 (any num)))
+      ;; " "
+      ;; (= 2 (any num)) ":" (= 2 (any num)) ":" (= 2 (any num)))
+    ;; tags, bookmarks, commit-id
+    (group
+      (*? nonl))
+
+    " "
+    ;; commit-id
+    (group
+      (+ (any hex)))
+    " "
+    ;; special states
+    (? (group (or (: "(empty) ")
+                "")))
+    (? (group (or (: "(no description set)" (* nonl))
+                "")))
+    ;; regular description
+    (* nonl)
+    line-end))
+
+(define-derived-mode vc-jj-log-view-mode log-view-mode "JJ-Log-View"
+  (require 'add-log) ;; We need the faces add-log.
+  ;; Don't have file markers, so use impossible regexp.
+  (setq-local log-view-file-re regexp-unmatchable)
+  (setq-local log-view-per-file-logs nil)
+  (setq-local log-view-message-re vc-jj--logline-re)
+  ;; Allow expanding short log entries.
+  (setq truncate-lines t)
+  (setq-local log-view-expanded-log-entry-function
+    'vc-jj-expanded-log-entry)
+  (setq-local log-view-font-lock-keywords
+    `((,vc-jj--logline-re
+        (1 'log-view-message)
+        (2 'change-log-list)
+        (3 'change-log-name)
+        (4 'change-log-date)
+        (5 'change-log-file)
+        (6 'change-log-list)
+        (7 'change-log-function)
+        (8 'change-log-function))))
+
+  (keymap-set vc-jj-log-view-mode-map "r" #'vc-jj-edit-change)
+  (keymap-set vc-jj-log-view-mode-map "x" #'vc-jj-abandon-change)
+  (keymap-set vc-jj-log-view-mode-map "i" #'vc-jj-new-change))
+
+
+(defun vc-jj-expanded-log-entry (revision)
+  (with-temp-buffer
+    (vc-jj--command-dispatched t 0 nil
+      "log"
+      "-r" revision
+      "-T" "builtin_log_detailed"
+      "--color" "never"
+      "--no-graph"
+      "--")
+    (buffer-string)))
+
+(defun vc-jj-previous-revision (file rev)
+  "JJ-specific version of `vc-previous-revision'."
+  ;; TODO: here and in `vc-jj-next-revision', check how (concat
+  ;; revision "-") works with merge commits; do we get multiple change
+  ;; ids in a multi-line string, and does this break users of these
+  ;; functions?
+  (if file
+      (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
+                                "-r" (concat ".." rev "-")
+                                "-T" "change_id.shortest()"
+                                "--" (vc-jj--filename-to-fileset file))
+    (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
+                              "-r" (concat rev "-")
+                              "-T" "change_id.shortest()"
+                              )))
+
+(defun vc-jj-next-revision (file rev)
+  "JJ-specific version of `vc-next-revision'."
+  (if file
+      (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
+                                "-r" (format "roots(files(\"%s\") ~ ::%s)" file rev)
+                                "-T" "change_id.shortest()")
+    (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
+                              "-r" (concat rev "+")
+                              "-T" "change_id.shortest()")))
+
+(defun vc-jj-get-change-comment (_files rev)
+  (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
+                            "-r" rev "-T" "description"))
+
+(declare-function log-edit-mode "log-edit" ())
+(declare-function log-edit-toggle-header "log-edit" (header value))
+(declare-function log-edit-extract-headers "log-edit" (headers string))
+(declare-function log-edit--toggle-amend "log-edit" (last-msg-fn))
+
+;; TODO: protect immutable changes
+(defun vc-jj-modify-change-comment (_files rev comment)
+  (let ((args (log-edit-extract-headers () comment)))
+    (vc-jj--command-dispatched nil 0 nil "desc" rev "-m" (pop args) "--quiet")))
+
+(defun vc-jj--reload-log-buffers ()
+  (and vc-parent-buffer
+    (with-current-buffer vc-parent-buffer
+      (revert-buffer)))
+  (with-current-buffer (revert-buffer)))
+
+(defun vc-jj-edit-change ()
+  (interactive)
+  (let ((rev (log-view-current-tag)))
+    (vc-jj--command-dispatched nil 0 nil "edit" rev "--quiet")
+    (vc-jj--reload-log-buffers)))
+
+(defun vc-jj-abandon-change ()
+  (interactive)
+  ;; TODO: should probably ask for confirmation, although this would be
+  ;; different from the cli
+  (let ((rev (log-view-current-tag)))
+    (vc-jj--command-dispatched nil 0 nil "abandon" rev "--quiet")
+    (vc-jj--reload-log-buffers)))
+
+(defun vc-jj-new-change ()
+  (interactive)
+  (let ((rev (log-view-current-tag)))
+    (vc-jj--command-dispatched nil 0 nil "new" rev "--quiet")
+    (with-current-buffer (revert-buffer))
+    (vc-jj--reload-log-buffers)))
+
 (defun vc-jj-root (file)
   "Return the root of the repository containing FILE.
 Return NIL if FILE is not in a jj repository."
@@ -433,7 +653,7 @@ For jj, modify `.gitignore' and call `jj untrack' or `jj track'."
   (let ((default-directory
          (if directory (file-name-as-directory directory)
            default-directory)))
-    (vc-jj-command nil 0 file "file" (if remove "track" "untrack") "--")))
+    (vc-jj--command-dispatched nil 0 file "file" (if remove "track" "untrack") "--")))
 
 (defun vc-jj-diff (files &optional rev1 rev2 buffer _async)
   "Display diffs for FILES between revisions REV1 and REV2."
@@ -451,7 +671,7 @@ For jj, modify `.gitignore' and call `jj untrack' or `jj track'."
       (erase-buffer))
     (apply #'call-process vc-jj-program nil buffer nil "diff" "--from" rev1 "--to" rev2 args)
     (if (seq-some (lambda (line) (string-prefix-p "M " line))
-                  (apply #'process-lines vc-jj-program "diff" "--summary" "--" files))
+                  (apply #'vc-jj--process-lines "diff" "--summary" "--" files))
         1
       0)))
 
@@ -459,9 +679,12 @@ For jj, modify `.gitignore' and call `jj untrack' or `jj track'."
   "Fill BUF with per-line commit history of FILE at REV."
   (let ((rev (or rev "@"))
         (file (file-relative-name file)))
-    (apply #'vc-jj-command buf 'async nil "file" "annotate"
-      (append (vc-switches 'jj 'annotate)
-        (list "-r" rev file)))))
+    ;; Contrary to most other jj commands, 'jj file annotate' takes a
+    ;; path instead of a fileset expression, so we append FILE to the
+    ;; unprocessed argument list here.
+    (apply #'vc-jj--command-dispatched buf 'async nil "file" "annotate"
+           (append (vc-switches 'jj 'annotate)
+                   (list "-r" rev file)))))
 
 (defconst vc-jj--annotation-line-prefix-re
   (rx bol
@@ -500,8 +723,7 @@ four groups: change id, author, datetime, line number.")
 
 (defun vc-jj-revision-completion-table (files)
   (let ((revisions
-         (apply #'process-lines
-                vc-jj-program "log" "--no-graph"
+         (apply #'vc-jj--process-lines "log" "--no-graph"
                 "-T" "self.change_id() ++ \"\\n\"" "--" files)))
     (lambda (string pred action)
       (if (eq action 'metadata)
@@ -565,5 +787,8 @@ the command to run, e.g., the semi-standard \"jj git push -c @-\"."
                         nil))))
   (vc-set-async-update buffer))
 
+;;;###autoload
+(add-to-list 'auto-mode-alist '("\\.jjdescription\\'" . log-edit-mode))
+
 (provide 'vc-jj)
 ;;; vc-jj.el ends here
